/*
The software in this package is distributed under the GNU General
Public License version 2 (with a special exception described below).

A copy of GNU General Public License (GPL) is included in this distribution,
in the file COPYING.GPL.

As a special exception, if other files instantiate templates or use macros
or inline functions from this file, or you compile this file and link it
with other works to produce a work based on this file, this file
does not by itself cause the resulting work to be covered
by the GNU General Public License.

However the source code for this file must still be made available
in accordance with section (3) of the GNU General Public License.

This exception does not invalidate any other reasons why a work based
on this file might be covered by the GNU General Public License.
*/
#ifndef PING_SCHEDULER_H
#define PING_SCHEDULER_H

#include <stdint.h>

#include <string>
#include <vector>

#include <boost/asio.hpp>
#include <boost/shared_ptr.hpp>

#include "link/linkstatus.h"
#include "host/hoststatus.h"
#include "host/pinger.h"
#include "host/pinginterval.h"
#include "host/pingprotocol.h"
#include "host/pingrotate.h"

//-----------------------------------------------------------------------------
// PingScheduler
//-----------------------------------------------------------------------------

/**
 * @brief This class is responsible to control whether to ping a host. It
 * schedules periodic pings to a host.
 * Scope: one object per host.
 */
class PingScheduler
{
public:
    PingScheduler(
            const IoServiceItem io_service,
            const std::string &network_interface,
            const std::string &destination_address,
            const uint16_t destination_port,
            const PingProtocolList &ping_protocol_list,
            const long ping_interval_in_sec,
            const int ping_fail_percentage_limit,
            const std::string &nameserver,
            const int max_address_resolution_attempts,
            const int ping_reply_timeout,
            const int resolved_ip_ttl_threshold,
            LinkStatusItem link_analyzer,
            const int first_delay
    );
    ~PingScheduler();

    void start_pinging();
    void stop_pinging();

private:
    //
    // Methods
    //

    void resolve_and_ping(const boost::system::error_code &error);
    bool resolve_address();
    void force_address_resolution();

    void ping();
    void ping_done_handler(const bool ping_success);
    void update_ping_interval();
    void update_ping_elapsed_time();

    //
    // Attributes
    //

    /// Timer to trigger the next ping
    boost::asio::deadline_timer NextPingTimer;
    /// Timer to trigger the next attempt to resolve addresses
    boost::asio::deadline_timer NextAddressTimer;
    /// Keeps track of the time when the last ping was send
    boost::posix_time::ptime TimeSentLastPing;
    /// Interval between each ping to the same host
    PingInterval PingIntervalInSec;
    /// Interval between attempts to resolve host address
    PingInterval AddressResolveIntervalInSec;
    /// Object responsible to evaluate the status of the host
    HostStatus HostAnalyzer;
    /// delay for very first ping to avoid lots of simultaneous pings
    int  FirstDelay;
    /// number of attempts at Address Resolution
    int AddressResolutionAttempts;
    int MaxAddressResolutionAttempts;
    /// flag whether any DNS lookup succeeded
    //  (comparing get_ip_count to 0 might violate assertion in dnsresolver.cpp)
    bool EverHadAnyIP;
    /// Internal boost pinger object
    PingRotateItem Ping;
};

//-----------------------------------------------------------------------------
// PingSchedulerItem
//-----------------------------------------------------------------------------

typedef boost::shared_ptr<PingScheduler> PingSchedulerItem;

//-----------------------------------------------------------------------------
// PingSchedulerList
//-----------------------------------------------------------------------------

typedef std::vector< PingSchedulerItem > PingSchedulerList;

#endif // PING_SCHEDULER_H
