// Copyright (c) 2003-2010 Christopher M. Kohlhoff
// Modifications (c) 2011 by Guilherme Maciel Ferreira / Intra2net AG
//
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)
#ifndef ICMPV6_PACKET_H
#define ICMPV6_PACKET_H

#include <stdint.h>

#include <istream>
#include <ostream>

#include <boost/asio.hpp>

#include "icmp/icmppacket.h"
#include "icmp/icmpv6header.h"
#include "icmp/icmpdata.h"
#include "icmp/icmptype.h"
#include "ip/ipv6header.h"

//-----------------------------------------------------------------------------
// Icmpv6Packet
//-----------------------------------------------------------------------------

/**
 * @brief This class represents the ICMP version 6 Packet.
 *
 * The ICMP Packet format is:
 *
 * @code
 * 0      3 4            11 12   15 16          23 24            31
 * +-------+---------------+--------------------------------------+      ---
 * |       |               |                                      |       ^
 * |version| traffic class |              flow label              |       |
 * |  (6)  |               |                                      |       |
 * +-------+---------------+-------+--------------+---------------+       |
 * |                               |              |               |       |
 * |        payload length         | next header  |   hop limit   |       |
 * |                               |              |               |       |
 * +---------------+---------------+--------------+---------------+       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                      source IPv6 address                     | IPv6 Header
 * |                                                              |   40 bytes
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * +--------------------------------------------------------------+       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                   destination IPv6 address                   |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       v
 * +---------------+---------------+------------------------------+      ---
 * |               |               |                              |       ^
 * |     type      |     code      |          checksum            |       |
 * |               |               |                              |       |
 * +---------------+---------------+------------------------------+       |
 * |                               |                              | ICMP Payload
 * |          identifier           |       sequence number        | (header +
 * |                               |                              |       data)
 * +-------------------------------+------------------------------+   8+ bytes
 * |                                                              |       |
 * |                      data (optional)                         |       |
 * |                                                              |       v
 * +-------------------------------+------------------------------+      ---
 * @endcode
 */
class Icmpv6Packet : public IcmpPacket
{
public:
    Icmpv6Packet();
    Icmpv6Packet(
            const Icmpv6Header &icmp_header,
            const IcmpData &icmp_data
    );
    virtual ~Icmpv6Packet();

    Ipv6Header get_ip_header() const;
    Icmpv6Header get_icmp_header() const;
    IcmpData get_icmp_data() const;

    virtual bool match_echo_reply(
            const uint16_t identifier,
            const uint16_t sequence_number,
            const boost::asio::ip::address &source_address
    ) const;
    virtual bool match_destination_unreachable(
            const uint16_t identifier,
            const uint16_t sequence_number,
            const boost::asio::ip::address &source_address
    ) const;

    virtual void print_echo_reply(
            const std::size_t &bytes_transferred,
            const boost::posix_time::ptime &time_packet_sent
    ) const;
    virtual void print_destination_unreachable() const;

    virtual ReadReturnCode read( std::istream &is );
    bool write( std::ostream &os ) const;


private:
    bool match(
            const Icmpv6Type type,
            const uint16_t identifier,
            const uint16_t sequence_number,
            const boost::asio::ip::address &source_address
    ) const;

private:
    /// The IP header.
    Ipv6Header IpHeader;
    /// The ICMP packet header.
    Icmpv6Header IcmpPayloadHeader;
    /// The ICMP packet payload (data).
    IcmpData IcmpPayloadData;
};

#endif // ICMPV6_PACKET_H
